// Javascript functions for WebDialog InputBox replacement
// wdAngleDim_settings.js
// John McClenahan, Steve Baumgartner
// MIT Licence
// v5.5 2016-12-21 ENGLISH


// **** IMPORTANT **** Set number of fields in your WebDialog ****
numberOfFields = 8;

//Initialise form status

formValid = null;

var use_comma = false;
// This function is called from Ruby only if SU uses comma separator
function lengths_use_comma() {
  use_comma = true;
}

//  -------------------------------------------------------------------------------------
// Set up tooltips to show user what types and range of values are accepted in each field
//
var i=1;

while (i <= numberOfFields) {
  var fieldID = "field" + i;
  var e = document.getElementById(fieldID);

  // Set up tooltip strings for non-SELECT fields to show required type, and if set, min and/or max values
  // Leave static 'title' in select field(s) as tooltip
  if (e.tagName != 'SELECT' ) {
    var minStr = '';
    if(isNumeric(e.min) && e.min !='') {minStr = " >= " + e.min;}
    var maxStr = '';
    if(e.max != "" && isNumeric(e.max)) {maxStr = "; <= " + e.max;}
    e.title = "Der Wert sollte vom Typ sein '" + e.placeholder + "'" + minStr + maxStr;
  }
  i++;
} // End while

function commaFormatPrompts() {
  var i=2;
  while(i <= numberOfFields - 1) {
    var fieldID = "field" + i;
    var e = document.getElementById(fieldID);
    new_title = e.title.replace(/(\d)\.(\d)/g, '$1' + ',' + '$2');
    e.title = new_title;
    var min = e.min.replace(/(\d)\.(\d)/g, '$1' + ',' + '$2');
    e.min = min;
    var max = e.max.replace(/(\d)\.(\d)/g, '$1' + ',' + '$2');
    e.max = max;
    var promptID = fieldID+"prompt";
    var p = document.getElementById(promptID);
    var prompt = p.innerHTML.replace(/(\d)\.(\d)/g, '$1' + ',' + '$2');
    p.innerHTML = prompt;
    i++;
  }
}

// --------------------------------------------------------------------------------------
// onblur for any field, checkField(field,type) is called.
// If input contents are valid type (and if relevant, range), then field text color is set to black
// Otherwise, text color is set to red, and focus is frozen on the field in error.
// Field text is all highlighted ready for overtyping to edit, and the warning message
//   color is set to match the field color: black for valid, red for error

function checkField(field, type) {

  var e = document.getElementById(field);
  var x = e.value;
  var min = e.min;
  var max = e.max;
  var warn = document.getElementById("warning");

  // If there's an invalid entry, hold the cursor in the field, and display warning
  switch(type) {

  case 'numeric':
    if (isNumeric(x)){
      //          x = parseFloat(x);
      if (inRange(x , min, max)) {
	e.style.color = "black";
	return true;
      } else {
	e.focus();
	e.style.color = "red";
	warn.visible = true;;
	return false;
      }
    }
  case 'integer':
    if (isInteger(x)){
      //          x = parseInt(x);
      if (inRange(x , min, max)) {
	e.style.color = "black";
	return true;
      } else {
	e.focus();
	e.style.color = "red";
	warn.visible = true;
	return false;
      }
    }

  case 'length':
    if (isImperialLength(x) || isMetricLength(x)) {
      e.style.color = "black";
      return true;
    } else {
      e.focus();
      e.style.color = "red";
      warn.visible = true;
      return false;
    }

  case 'checkbox':
    return true;

  case 'filename':
    if (isFilename(x)) {
      e.style.color = "black";
      return true;
    } else {
      e.focus();
      e.style.color = "red";
      warn.visible = true;
      return false;
    }

  default:
    e.style.color = "black";

  } //End switch

} // End checkField function

// onfocus for any field, resetField(field) is called.
function resetField(field){
  var e = document.getElementById(field);
  var warn = document.getElementById("warning");
  if(formValid == null || formValid == true) {
    e.style.color = "black";
    var l = document.getElementById(field +"label").innerHTML;
    //       var warn = document.getElementById("warning");
    //     var info = document.getElementById("info");
    warn.innerHTML = '';
    //       warn.style.color = "black";
    info.innerHTML = e.title;
    info.style.color = "black";
    if (field != "field1") {
      e.select(); // Allow immediate overtyping.  No select method on <select> elements, though.
    }
    return true;
  } else {
    e.focus(); // Keep focus on error field
    if (field != "field1") {
      e.select(); // Allow immediate overtyping.  No select method on <select> elements, though.
    }
    return false;
  }
}

// --------------------------------------------------------------------------------------
// When OK button is clicked, this function checks all fields again, and if any are invalid
//  (text color set to red), disables OK button and sets focus on the first invalid entry

function button_clicked() {
  send_values("OK");
}

function send_values(flag) {
  // Ok button has been clicked, or Enter key pressed to trigger it
  var i=1;
  var warn = document.getElementById("warning");
  var info = document.getElementById("info");
  formValid = true; // until shown otherwise by checkField
  // Loop through fields, check if any have invalid input,
  while (i <= numberOfFields) {
    var field = "field" + i;
    var lbl = document.getElementById(field +"label").innerHTML;
    e = document.getElementById(field);
    var type = e.placeholder;
    var check = null;
    if (e.tagName == 'SELECT') {// Skip checking any SELECT fields
      check = true;
    } else {
      check = checkField(field, type);
      formValid = check;
    } // End if SELECT

    if (!check){ // Error in a field
      warn.style.color =  "red";
      warn.innerHTML = "Sie haben einen ungültigen Wert in  " + lbl + ". Bitte korrigieren Sie und versuchen Sie es erneut <br />";
      info.style.color = "red";
      info.innerHTML = e.title; // the title shows the valid field type (and range if any)
      e.focus();
      e.select();
      return; // stop running further after first error found
    } else {
      warn.style.color =  "black";
      warn.innerHTML = "";
      info.style.color = "black";
      info.innerHTML = e.title; // the title shows the valid field type (and range if any)
    } // End if check
    i++;
  } // End while checking

  //---------------------------------------------------------------------------
  if (formValid) {
    info.innerHTML = ''; // Reset info display
    // If all ok, compile result string to send to Ruby
    //  All fields validated. Assemble result string from field values.
    var result = '';

    i=1;
    while (i <= numberOfFields) {
      var field = "field" + i;

      var lbl = document.getElementById(field +"label").innerHTML;
      e = document.getElementById(field);
      if (e.placeholder == 'checkbox') {
        if (e.checked == true) {
          result += 'yes' + '|';
        } else {
          result += 'no' +'|';
        }
      } else {
        result += e.value + '|';
      }
      i++;
    } // End while writing results

    // Convert single and double quotes (foot and inch marks) to HTML encodings
    result = result.replace(/"/g,'&quot;');
    result = result.replace(/\'/g,"&apos;");

    // Remove trailing |
    result = result + flag;

    result = encodeURIComponent(result);
    sendToRuby( 'get_inputs', result );
  } // End if formValid
} // End function send_values

// --------------------------------------------------------------------------------------
// The following three JS functions, and the Ruby callback functions get_inputs() and set_values()
//  work together to pass messages back and forth between this WebDialog page's input fields and Ruby
function sendToRuby( cb_name, msg ) {
  fake_url = 'skp:' + cb_name + '@' + msg;
  window.location = fake_url;
}// End function sendToRuby

function load_values() {
  //   Loads values sent from Ruby by running its callback function 'set_values'
  sendToRuby('set_values', "" );
} // End function load_values

// tell Ruby to replace the current values with saved defaults
function load_defaults() {
  sendToRuby('load_defaults', "");
}

// tell Ruby to save the current values as defaults.
// This needs to first update Ruby's values because the user
// will naturally expect to see the currently displayed values
// become the defaults, not what was previously saved to Ruby!
function save_defaults() {
  send_values("defaults");
}

// second stage of save, to account for asynchronous JS on Mac
function save_defaults_2() {
  sendToRuby('save_defaults', "");
}

function set_inputs(msg) {
  //   Load values sent from Ruby callback function set_values()
  // Replace html apostrophe and quote by ft and inch marks (single and double quotes)
  msg = msg.replace(/&apos;/g, "'");
  msg = msg.replace(/&quot;/g, '"');
  currentValues = msg.split("|");
  // Loop through fields and split values to fill in fields
  var i=1;

  while (i<= numberOfFields - 1) {
    var fieldID = "field" + i;
    document.getElementById(fieldID).value = currentValues[i-1];
    i++;
  } // End while
  // For last (checkbox) field8, convert from 'yes' or 'no' to checked = true or false
  if (currentValues[7] == 'yes') {
    document.getElementById("field8").checked = true;
  }else{
    document.getElementById("field8").checked = false;
  }

} // End function setInputs

// --------------------------------------------------------------------------------------



function isNumeric(str){
  var re1;
  if(use_comma) {
    re1 = /^-?\d*([\,]\d*)?$/;
  } else {
     re1 = /^-?\d*([\.]\d*)?$/;
  }
  return re1.test(str);
}

function isInteger(str) {
  var re2 = /^\-?([0-9]\d*)$/;
  return re2.test(str);
}

function isImperialLength(str) {
  var re3;
  if(use_comma) {
      var re3 = /^(~ )?\-?(\d*([\,]\d*)?\'?)? *\d*([\,]\d*)? ?(\d+\/(2|4|8|16|32|64))?(\")?$/;
  } else {
      var re3 = /^(~ )?\-?(\d*([\.]\d*)?\'?)? *\d*([\.]\d*)? ?(\d+\/(2|4|8|16|32|64))?(\")?$/;
  }
  //var re3 = /^(~ )?\-?(\d*([\.\,]\d*)?\'?)? *\d*([\.\,]\d*)? ?(\d+\/(2|4|8|16|32|64))?(\")?$/;
  return re3.test(str);
}

function isMetricLength(str) {
  var re4;
  if(use_comma) {
    re4 = /^(~ )?\-?\d*([\,]\d*)?(m|cm|mm)?$/;
  } else {
    re4 = /^(~ )?\-?\d*([\.]\d*)?(m|cm|mm)?$/;
  }
  //var re4 = /^(~ )?\-?\d*([\.\,]\d*)?(m|cm|mm)?$/;
  return re4.test(str);
}

function isFilename(str) {
  // Put in regex expression to limit valid characters - for example, for component or group name
  // limit to valid filename characters since components may be Saved As... files.
  // Original regex to ALLOW only valid characters  var RE5 = /^[0-9a-zA-Z ... ]+$/ ;

  // Can't get a Replacement regex to EXCLUDE invalid characters though something like this
  //   SHOULD work (but this doesn't) /(^[\|&;\$%@"<>\(\)\+,])/;
  var re5 = /^[0-9a-zA-Z ._]+$/ ;
  return re5.test(str);
}

function inRange(x, min, max) {
  var val = Number(x.replace(',','.'));
  min = Number(min);
  max = Number(max);
  if ((!isNumeric(min) || val >= min) && (!isNumeric(max) || val <= max ))  {
    return true;
  } else {
    return false;
  }
}



// This code responds to Enter key press and sends current values to Ruby
var body = document.getElementsByTagName('body')[0];
body.onkeydown = function (e) {
  if (e.keyCode === 13) {  //enter key code
    send_values("click");
  }
}

function openPage(page) {
  sendToRuby('show_help', page);
}
