# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright © 2016 Stephen Baumgartner and John McClenahan <slb.jwm@gmail.com>
# v5.5 beta 2016-12-21
# 3D text on dimensioned angle plane originally suggested and partially coded
#   by John McClenahan 2016

# Permission to use, copy, modify, and distribute this software for
# any purpose and without fee is hereby granted, provided the above
# copyright notice appears in all copies. Subject to MIT licence terms
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Extensively changed and adapted from Steve Baumgartner's earlier plugin
# Angular Dimension v4.
# This was extensively modified and enhanced from dim_angle.rb,
# Copyright 2005, Didier Bur, which, in turn, was based on the demo
# rectangle.rb by @Last Software.
# LanguageHandler added by Mario Chabot 2016. www.formation-sketchup.quebec

#-----------------------------------------------------------------------------

## Version history
#   v6.1      Added Swedish and Russian translations
#   v6.0      Completed and checked translations for five latin-alphabet languages
#               Adapted Measurement box display to show when User radius has been set, blank othersise
#   v5.6a-e beta Added two variants of dimension text and arc when small angle won't let both arrows
#             and text fit between dimension lines.
#   v5.5 beta Fixes to issues with square mark option, changed so doesn't draw then hide text-type
#            dimension when using square for right angle.
#   v5.4 beta Added option for square mark or angle for 90° angles
#   v5.3 beta Fixes to get comma separator working properly in Help page as well as Settings page
#               and to update version history comments and version numbering throughout
#   v5.2 beta Fix for comma or dot as decimal separator not working properly for lengths
#   v5.1 beta Edits to allow comma or dot as decimal separator in numbers and lengths
#   v5.0 beta a-d Edits to stop IE blocking JS scripts from running, and to hide Close button on Windows.
#   v5.0beta Minor edits to default values, help text and settings html file for consistency
#   v4.26  - Final packaging for beta test, including security signature for new versions of SU
#             and patches for SU 8 compatibility.
#   v4.25a - Tweaks to rotate arrowhead to line up with arc, not set at 90° to witness lines,
#             and to widen text_height_ and text_width_gap_angle a fraction more
#   v4.25 - Incorporated SLB changes to menu, classdata, and settings.rb files,
#             and JWM edit to make arc segments proportional to angle.
#           Changed settings.html page to send changed values of arrow_style immediately, and
#             changed text box size from 8 to 14
#           Amended copyright notice
#   v4.24 - JWM changed logic.rb coding to place text outside dimension arc if text won't fit,
#             perpendicular or parallel to angle bisector
#   v4.23 -  SLB fixed Web Dialog being allowed to open multiple times
#   v4.22b - JWM Fixed warnings for extra space before JS function parentheses, and/or missing parens
#             NOT fixed NoMethodError for 'require_relative' still somehow being found
#   v4.22 - JWM added logic to use require instead of require relative in logic.rb for older versions,
#             and to add 'unless defined?' to prevent warnings when PATH constants redefined in loader file
#   v4.21 - JWM tweaked the Regexes to allow for approximate values being return from Ruby (for example,
#             if entered units were metric in a model with architectural/imperial units set, or vice versa,
#             or where entered length is more precise than model precision setting), and changed
#             inRange function to use parseFloat instead of Numeric to check x, min and max values.
#           Also edited defaults, and help pages to match, so text_scale is now 0.10 not 0.09,
#             and arc_segments is set to 8 not 12. Move Help page [Close] button to top right,
#             since it often disappeared at bottom and needed scroll down to be seen.
#           Cosmetic changes to Settings page to emphasize OK button, and improve appearance
#             of Load and Save defaults buttons.
#   v4.20 - SLB fixed a lot of the code, including getting dialogue to pop up, and re-did the JS<-->Ruby
#             callbacks
#   v4.19 - Almost working dialgoue, but it wouldn't pop up when ctrl key pressed
#   v4.18 - Began to convert InputBox to wdInputBox in classdata.rb file- uses Web Dialogue instead,
#           with error checking, comments, and a help file
#   v4.17 - Changed order of inputs in InputBox in classdata.rb
#           Moved the JS script into a separate .js file
#           Moved and renamed wdInputBox files into JWM_draw_angle_dim folder
#   v4.16 - Fixed missing black and reversed face for dot arrowhead.
#           Lifted heel of arrows and all of dot 'arrowhead' to prevent Z-fighting
#           Changed default Text Scale from 0.05 to 0.09
#           Added Arc Radius and Text Scale to user InputBox in classdata.rb
#           Changed setting from -1 to 0.0 for Text Height and Arc Radius to use text_scale and picked radius
#   v4.15 - code clean-up and refactoring by SLB - lost anti-Z-fighting changes from 4.14
#   v4.14 - Filled text black, and lifted it 0.005 * @radius off dimension plane to reduce z-fighting if on face.
#          Have also filled closed arrow black, but left it on surface. Lifting it disp lays poorly from some angles
#          - noticeably out of line with arc end. Try lifting heel of arrow instead
#          Could perhaps stop arc short at middle of arrowhead for closed arrow only?
#          Added right angle lines, and hid arcs, arrowheads and text for right angles. Not sure how to scale
#          the lines, and/or whether to keep text displayed instead of hidden
#   v4.13 - Arrowhead now selectable simply by changing value of @arrow_style in initialize function
#             (but may need to restart SU?)
#   v4.12 - Outside angles working properly, including text angle
#   v4.11 - Outside angle arcs and arrowheads correctly drawn, and text correctly placed
#           but incorrectly oriented wrt screen
#   v4.10 - Text that won't fit in dimension angle correctly positioned and oriented,
#           AND you can choose which side of the angle
#           external text for small angles is put - it goes at the first picked end
#   v4.09 - Text that won't fit in dimension angle now correctly positioned and oriented at start
#           though I can't understand quite why it needs 90 deg different rotation from text inside angle
#   v4.08 - Moved make_arrowhead into a function, and reordered some of the steps for future use
#   v4.07 - tests whether whole dimension, only text, or neither, will fit inside dimension lines
#           and puts arrows and if necessary text, outside dimension lines.
#           Have still to rename vector to text position, to get the orientation of text correct on screen after move to vertex
#
#   v4.06 - senses dimension angle bisector relative to screen orientation,
#          and rotates text accordingly in 90 degree increments.
#          Removed text fill, which caused Z-fighting. Removed test display of angle bisector
#
#   v4.05 - now sensing direction of vertex relative to screen - using view.screen_coords, and rotating text 180 degrees if nec.
#   v4.04 - partially working detection of direction of vertex, to rotate text if needed
#   v4.03 - working with SLB transformation matrix, and three arrow styles implemented. Text can still be upside down
#   v4.02 - partially working
#   v4.01 - SLB code adapted for namespace change to JWM - works except not finding Images and Resources
#          but works if those folders are placed in Plugins in a folder JWM_draw_angle_dim
##---JWM

target_file1 = 'drawAngleDim_classdata.rb'
target_file2 = 'drawAngleDim_callbacks.rb'

require File.join(SLB_JWMPlugins::AngularDim::EXTPATH, target_file1)
require File.join(SLB_JWMPlugins::AngularDim::EXTPATH, target_file2)

module SLB_JWMPlugins
  module AngularDim

    class DrawAngleDimTool
      ########################
      # Instance initializer
      #
      def initialize
        # @ip is the InputPoint from the current pick
        @ip = Sketchup::InputPoint.new
        # @ip1 is the InputPoint from the previous pick
        # this is used as reference when @ip picks, which, in principle, should
        # cause the inference engine to favor points on the edge containing the
        # previous pick
        @ip1 = Sketchup::InputPoint.new

        @drawn = false

        # @pts[0] = first picked point
        # @pts[1] = picked vertex
        # @pts[2] = second picked point
        @pts = []

        # @state = 0 => waiting for first pick
        # @state = 1 => waiting for vertex pick
        # @state = 2 => waiting for second pick
        # @state = 3 => have three points, ready to draw
        @state = 0

        # radius for drawn arc can be calculated by default or input via VCB
        @radius = 0

        # Initialize refs to component definitions for the available styles.
        # If already existing in the model, reuse the definition.  Otherwise
        # wait until a style is actually used before creating it (these initializations
        # will get nil).  Definitions are held by the model, so they automatically
        # persist across invocations of the tool.
        @dim_angle_aro_open = Sketchup.active_model.definitions["dim_angle_aro_open"]
        @dim_angle_aro_closed = Sketchup.active_model.definitions["dim_angle_aro_closed"]
        @dim_angle_aro_slash = Sketchup.active_model.definitions["dim_angle_aro_slash"]
        @dim_angle_aro_dot = Sketchup.active_model.definitions["dim_angle_aro_dot"]
        @dim_angle_aro_none = Sketchup.active_model.definitions["dim_angle_aro_none"]

        # tab key toggles between drawing inside and outside angle dimension
        @inside = true
        @cursor = IN_CURSOR

        @wd = nil
        @wdhelp = nil
      end # initialize

      def drawn?
        @drawn
      end

      def state
        @state
      end

      #-----------------------------------------------------------------------------
      def show_status
        Sketchup::set_status_text((DangLH['Arc radius']), SB_VCB_LABEL)
        user_radius = self.class.user_radius
        if @state == 0
          # may arrive here with a newly started Tool, so need to set the prompt
          Sketchup::set_status_text(DangLH['Point 1: first end of measured angle'])
        end

        # put the current radius into the VCB
        if !user_radius.nil? && user_radius != 0
          # have a valid user radius - show it
          Sketchup::set_status_text Sketchup.format_length(user_radius), SB_VCB_VALUE
        elsif @state == 0
          # no valid user radius and point-based radius not known yet -
          # show blank
          Sketchup::set_status_text "", SB_VCB_VALUE
        else
          # no valid user radius, but point-based should be known because not state 0 -
          # show the calculated radius
          Sketchup::set_status_text Sketchup.format_length(@radius), SB_VCB_VALUE
        end
      end # def show status

      #-----------------------------------------------------------------------------
      # clear out all saved data from previous operation
      def reset
        @pts = []
        @state = 0

        @ip1.clear
        @drawn = false
        UI.set_cursor(@cursor)
        show_status
      end # def reset

      # #-----------------------------------------------------------------------------
      def set_current_point(x, y, view)
        case @state
        when 0
          # capture first end point
          @pts[0] = @ip.position
          # and cancel user's ability to rescale last drawn angle dim
          @drawn = false
        when 1
          # capture vertex point
          @pts[1] = @ip.position
          # tell the user what radius arc will be drawn, so they can change via VCB if desired
          length = @pts[0].distance @pts[1]
          if(self.class.user_radius > 0)
            @radius = self.class.user_radius
          else
            @radius = length
          end
        when 2
          # capture third point
          @pts[2] = @ip.position
        end
        show_status
      end

      #-----------------------------------------------------------------------------

      # Fetch the component definition for the specified arrowhead style.
      # The various @dim_angle_aro_... attributes are initialized to nil
      # and only populated with a definition when the style is first used.
      # This avoids cluttering a model with component definitions for
      # unused arrowhead styles.
      # The arrowhead geometry in the component definition is drawn
      # at the origin and will need transformation for correct orientation,
      # location, and scale when a new instance is created.
      def get_arrowhead_definition(arrow_style)
        # arrow_style - text: "closed", "open", "slash",  "dot" or "none"

        # Draw an arrowhead component
        # Lift heel of arrowhead to avoid Z-fighting if on a face
        arrow_points = Array.new ;
        arrow_points[0] = [1.0, -0.1667, 0.01]
        arrow_points[1] = ORIGIN ; # "ORIGIN" is a SU provided constant
        arrow_points[2] = [1.0,  0.1667, 0.01]

        case arrow_style # Set in initialize as closed, open, line, dot or none

        when "closed"
          if @dim_angle_aro_closed.nil? || @dim_angle_aro_closed.deleted?
            @dim_angle_aro_closed = Sketchup.active_model.definitions.add("dim_angle_aro_closed")

            arrow_face = @dim_angle_aro_closed.entities.add_face(arrow_points)
            # If the  blue face is pointing up, reverse it.
            arrow_face.reverse! if arrow_face.normal.z < 0  # flip face to up if facing down
            # Fill arrowhead
            arrow_face.material = "black"
          end
          arrowhead = @dim_angle_aro_closed
        when "open"
          if @dim_angle_aro_open.nil? || @dim_angle_aro_open.deleted?
            @dim_angle_aro_open = Sketchup.active_model.definitions.add("dim_angle_aro_open")
            arrow_lines = @dim_angle_aro_open.entities.add_edges arrow_points
          end
          arrowhead = @dim_angle_aro_open
        when "slash"
          if @dim_angle_aro_slash.nil? || @dim_angle_aro_slash.deleted?
            @dim_angle_aro_slash = Sketchup.active_model.definitions.add("dim_angle_aro_slash")
            arrow_points[0] = [0.5, -0.5, 0]
            arrow_points[1] = ORIGIN
            arrow_points[2] = [-0.5,  0.5, 0]
            arrow_lines = @dim_angle_aro_slash.entities.add_edges arrow_points
          end
          arrowhead = @dim_angle_aro_slash
        when "dot"
          if @dim_angle_aro_dot.nil? || @dim_angle_aro_dot.deleted?
            @dim_angle_aro_dot = Sketchup.active_model.definitions.add("dim_angle_aro_dot")
            # Lift dot by a small distance to avoid Z-fighting on a face
            arrow_dot = @dim_angle_aro_dot.entities.add_circle([0,0,0.01], Z_AXIS, 0.1667, 12)
            arrow_face = @dim_angle_aro_dot.entities.add_face(arrow_dot)
            # If the  blue face is pointing up, reverse it.
            arrow_face.reverse! if arrow_face.normal.z < 0
            # Fill arrowhead
            arrow_face.material = "black"
          end
          arrowhead = @dim_angle_aro_dot
        when "none"
          if @dim_angle_aro_none.nil? || @dim_angle_aro_none.deleted?
            @dim_angle_aro_none = Sketchup.active_model.definitions.add("dim_angle_aro_none")
            # have to have something in the component to allow it to be created (I think)
            # You can turn guides off, or maybe I can make this hidden?
            arrow_lines = @dim_angle_aro_none.entities.add_cpoint arrow_points[1]
          end
          arrowhead = @dim_angle_aro_none
        else
          UI.messagebox "Sorry that #{arrow_style} is not a valid arrow type. \n" +
                        "Please edit the script and initialize @arrow_style"
          arrowhead = nil
        end #case

        # return value
        arrowhead
      end # get_arrowhead_definition

      #-----------------------------------------------------------------------------
      # Returns a two-element array of width and height needed for the text,
      #   with an arbitrary bit extra for to space text from arcs
      def space_for_text(radius, text_bb_width, text_bb_height)
        space_for_text_width = 1.05 *(2.0 * Math::atan(0.5*text_bb_width/(@radius - 0.5*text_bb_height)))
        space_for_text_height = 1.05 * (2.0 * Math::atan(0.5*text_bb_height/(@radius - 0.5*text_bb_width)))
        return space_for_text_width, space_for_text_height
      end

      #-----------------------------------------------------------------------------
        # Draw the angular dimension on the model after all inputs collected
      def draw_angle_dim
        model = Sketchup.active_model

        ##  Basic Geometry Analysis

        # make sure we are using the user's selected radius regardless of
        # which state this object was in when it was selected.
        if(self.class.user_radius > 0)
          @radius = self.class.user_radius
        end # if

        # make vectors from the user's input points.  These vectors point
        # out from the angle vertex along the two edges.  They are needed by
        # some subsequent methods that require directions for offsets, not
        # points
        vec1 = @pts[1].vector_to @pts[0]
        vec2 = @pts[1].vector_to @pts[2]

        # trap degenerate cases
        if(vec1.length == 0 or vec2.length == 0)
          UI.messagebox(DangLH['You must select three distinct points!'])
          reset
          return
        end

        if(vec1.parallel? vec2)
          UI.messagebox(DangLH['The selected lines are parallel, angle is 0 or 180°'])
          reset
          return
        end

        # scale the vectors to extend the radius a little.  This will affect
        # the drawn edges.  The value is arbitrary - change @dim_line_scale
        # via the Settings dialogue if another look is desired
        vec1.length = @radius * self.class.dim_line_scale
        vec2.length = @radius * self.class.dim_line_scale

        # the angle bisector vector - used for placing the text
        bisector = (vec1+vec2)

        # Corresponding points at ends of bisector vector
        edge_bisector = []
        edge_bisector[0] = @pts[1]
        edge_bisector[1] = @pts[1].offset bisector

        # find the angle between the vectors and the normal to their plane
        # this calculation should not explode, since we trapped the case of
        # parallel above (angle = 0 or 180), but the calculation of the normal
        # will be numerically unstable when angle is very small or very close to
        # 180.  I (SLB) have tested at 0.1 degrees without problems.
        angle = vec1.angle_between vec2
        complement = 360.degrees - angle
        bisector_angle = 0.5 * angle

        # Work out which quadrant the angle bisector appears to be in screen view and set text rotation accordingly
        # Four cases, one for each quadrant (each with two variants dealt with later):
        @quadrant = which_quadrant(edge_bisector)

        quadrants = {"northerly"=> 0, "easterly" => 1, "southerly" => 2, "westerly" => 3}
        @text_rotn = bisector_angle + (quadrants[@quadrant] - 1) * 90.degrees

        # Calculate angle needed to accommodate arrowhead and minimum arc length

        arrow_atan = Math::atan(self.class.arrow_scale)

        case self.class.arrow_style
        when "closed", "open"
          @arrowhead_angle = 2.0 * arrow_atan
        when "slash"
          @arrowhead_angle = 1.5 * arrow_atan
        when "dot", "none"
          @arrowhead_angle = arrow_atan
        end #case

        # Set text depending on whether inside or outside angle was chosen
        text = Sketchup.format_angle(angle) + "°"
        text2  = Sketchup.format_angle(complement) + "°"

        ## Adjust direction of normal, and order of vec1, vec2, in relation to view angle
        ## so the dimension goes into the picked points the right way round
        normal = (vec1 * vec2).normalize

        camera = Sketchup.active_model.active_view.camera
        view_dir = camera.direction
        dot = normal.dot view_dir
        ccw = dot > 0
        if ccw
          normal.reverse!
          vec1,vec2 = vec2, vec1
        end

        ## Start drawing

        # this enables undo of the whole operation as a unit
        model.start_operation "Angular Dimension"

        # create a new group in which we will draw our dimension entities
        model_ents = model.active_entities
        group = model_ents.add_group

        if @inside
          group.name = "Angular Dimension (" + text + ")"
        else
          group.name = "Angular Dimension (" + text2 + ")"
        end

        ents = group.entities

        ##  Put in angle delimiter ('witness') lines at origin
        ents.add_edges [self.class.dim_line_scale*@radius, 0, 0], ORIGIN
        ents.add_edges ORIGIN, [self.class.dim_line_scale*@radius*Math::cos(angle),
                                self.class.dim_line_scale*@radius*Math::sin(angle),
                                0]

        # Handle special case of interior right angle marked with a square

        # Test if the dimension angle `text` starts with 90,
        #   then we'll be using SU's own precision to test for 90 degrees
        # The text will start ~ 90 if SU doesn't think it is exactly 90°.
        # And text2 (outside angle) can never start with 90

        if @inside && self.class.square90 == "yes" && text[0..1] == "90"
          # previous test was # angle <= 90.01.degrees && angle >= 89.99.degrees

          # Draw two straight lines at right angles at vertex
          # Scale the square down so as not to be overwhelming - 0.333 is arbitrary
          ra_scale = 0.333 * @radius
          right_angle = []
          right_angle[0] = [ra_scale,0,0]
          right_angle[1] = [ra_scale,ra_scale,0]
          right_angle[2] = [0,ra_scale,0]
          ents.add_edges right_angle
        else
          # not a right angle or not square90 mode, proceed to draw arrows and text

        ## Scale the arrowheads to desired size
          arrow_size_scale = Geom::Transformation.scaling self.class.arrow_scale * @radius

        ## Set text size as a fraction of @radius (that's the default, loaded from classdata.rb).
          # Or user may have set fixed height (see below)
          @text_scale = self.class.text_scale

          if !@inside
            text = text2
          end

          if self.class.user_text_height > 0.0
            # the user has input a specific text height
            text_height = self.class.user_text_height
          else
            # use the text scale and radius to define text height
            text_height = @text_scale * @radius
          end

        ## Draw angle dimension text in 3D text, inside a group, at the origin
          # Parameters are string, alignment, font name, is_bold (Boolean), is_italic (Boolean),
          #   letter_height, tolerance, z, is_filled (Boolean), extrusion
          text_group = ents.add_group
          t = text_group.entities.add_3d_text text, TextAlignLeft, "Arial", false, false, text_height, 0, 0, true, 0.0
          # Colo(u)r the text black
          #   (optional - can cause Z-fighting in display, minimised by giving it a small Z offset below)
          text_group.material = "black"

          # Find the centre, width and height of the text group from its bounding box
          #   (bb_height may be bigger than letter_height)
          text_bb_center = text_group.local_bounds.center
          text_bb_width = text_group.local_bounds.width
          text_bb_height = text_group.local_bounds.height

        ## Calculate angular gaps needed to accommodate width or height of text
        #    - space_for_text adds a little for spacing
          # Note: this has two elements - [0] is for use with text width in N or S quadrant
          #   element [1] is for use in E or W quadrant
          text_gap_angle = space_for_text(@radius, text_bb_width, text_bb_height)

          ## Interior angle mode ########################################

          if(@inside)
            ## Draw arcs according to text and arrow size and placement

            # Angle subtended by arrowhead defined by arrow scale and style
            # We need that to be able to rotate the arrowheads to line up with start and end of arc
            arrow_asin = Math::asin(0.5*self.class.arrow_scale)

            # Number of segments per 90° to use for dimension arcs
            segments = (self.class.arc_segments *  (angle/90.degrees)).to_i

            # Set up entities to receive arcs
            arcs_group = ents.add_group
            arc_ents = arcs_group.entities

            ## Calculate where to put arcs and how to rotate arrowheads for different cases
            case @quadrant
            when "northerly", "southerly"
              # Uses text width to adjust arc spacing
              if 2 * @arrowhead_angle + text_gap_angle[0] < angle
                #puts "N-S Dimension arrows and text will fit inside witness lines"
                @text_radius = @radius
                arc1 = arc_ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius, 0,
                        0.5*(angle - text_gap_angle[0]), segments
                arc2 = arc_ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius,
                        0.5*(angle + text_gap_angle[0]), angle, segments
                arrow1_rotn = 90.degrees + arrow_asin
                arrow2_rotn = angle - 90.degrees - arrow_asin

              elsif text_gap_angle[0] < angle
                #puts "N-S Text will fit inside witness lines, arrows go outside"
                @text_radius = @radius
                arc1 = arc_ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius, 0,
                         - @arrowhead_angle, segments
                arc2 = arc_ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius,
                        angle + @arrowhead_angle, angle, segments
                arrow1_rotn = - (90.degrees + arrow_asin)
                arrow2_rotn = angle + 90.degrees + arrow_asin

              elsif 2 * @arrowhead_angle < angle
                #puts "N-S Text has to go outside radius"
                @text_radius = @radius * self.class.dim_line_scale + 0.5 * text_bb_height
                # Draw complete arc (in two parts) inside witness lines, put text outside the arc
                arc1 = arc_ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius, 0,
                         0.5*@arrowhead_angle, segments
                arc2 = arc_ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius,
                         0.5*@arrowhead_angle, angle, segments
                arrow1_rotn = 90.degrees + arrow_asin
                arrow2_rotn = angle - 90.degrees - arrow_asin
              else
                  # Draw arcs outside witness lines, as well as text
                  @text_radius = @radius * self.class.dim_line_scale + 0.5 * text_bb_height
                  arc1 = arc_ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius, 0,
                          - @arrowhead_angle, segments
                  arc2 = arc_ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius,
                          angle + @arrowhead_angle, angle, segments
                  arrow1_rotn = - (90.degrees + arrow_asin)
                  arrow2_rotn = angle + 90.degrees + arrow_asin
              end
            when "easterly", "westerly"
              # Uses text height to adjust arc spacing
              if 2 * @arrowhead_angle + text_gap_angle[1] < angle
                #puts " E-W Dimension arrows and text will fit inside witness lines"
                @text_radius = @radius
                arc1 = arc_ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius, 0,
                        0.5*(angle - text_gap_angle[1]), segments
                arc2 = arc_ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius,
                        0.5*(angle + text_gap_angle[1]), angle, segments
                arrow1_rotn = 90.degrees + arrow_asin
                arrow2_rotn = angle - 90.degrees - arrow_asin

              elsif text_gap_angle[1] < angle
                #puts "E-W Text will fit inside witness lines, arrows go outside"
                @text_radius = @radius
                arc1 = arc_ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius, 0,
                          - @arrowhead_angle, segments
                arc2 = arc_ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius,
                          angle + @arrowhead_angle, angle, segments
                arrow1_rotn = - (90.degrees + arrow_asin)
                arrow2_rotn = angle + 90.degrees + arrow_asin

              elsif 2 * @arrowhead_angle < angle
                #puts "E-W Text has to go outside radius"
                @text_radius = @radius * self.class.dim_line_scale + 0.5 * text_bb_width
                # Draw complete arc (in two parts) inside witness lines, put text outside the arc
                arc1 = arc_ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius, 0,
                         0.5*@arrowhead_angle, segments
                arc2 = arc_ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius,
                         0.5*@arrowhead_angle, angle, segments
                arrow1_rotn = 90.degrees + arrow_asin
                arrow2_rotn = angle - 90.degrees - arrow_asin
              else
                # Draw arcs outside witness lines, as well as text
                @text_radius = @radius * self.class.dim_line_scale + 0.5 * text_bb_width
                arc1 = arc_ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius, 0,
                        - @arrowhead_angle, segments
                arc2 = arc_ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius,
                        angle + @arrowhead_angle, angle, segments
                arrow1_rotn = - (90.degrees + arrow_asin)
                arrow2_rotn = angle + 90.degrees + arrow_asin
              end
            end # case @quadrant

          ## Calculate arrow rotation transforms to line arrowheads up with arcs
          arrow1_rotate = Geom::Transformation.rotation ORIGIN, Z_AXIS, arrow1_rotn
          arrow2_rotate = Geom::Transformation.rotation ORIGIN, Z_AXIS, arrow2_rotn

          # Calculate moves to put arrowheads at start/end of arcs
          arrow1_move = Geom::Transformation.translation ORIGIN.vector_to(arc1[0].start.position)
          arrow2_move = Geom::Transformation.translation ORIGIN.vector_to(arc2[-1].end.position)

          # Get the chosen style of arrowhead to place in position
          dim_angle_arrow = get_arrowhead_definition(self.class.arrow_style)

          # Combine transformations to insert an arrowhead at start and end of arcs
          arrow1 = arc_ents.add_instance dim_angle_arrow, arrow1_move*arrow1_rotate*arrow_size_scale
          arrow2 = arc_ents.add_instance dim_angle_arrow, arrow2_move*arrow2_rotate*arrow_size_scale


          #-----------------  Place text
          #  Move the center of the text to the angle bisector, @text_radius from the vertex...
          text_center = Geom::Point3d.new [@text_radius*Math::cos(bisector_angle),
              @text_radius*Math::sin(bisector_angle),0]
          # Move text to center it on text_center,
          #   and lift text slightly above dimension plane, to avoid z-fighting if on a face
          text_posn = text_center - text_bb_center + [0, 0, 0.001*@radius]
          text_group.move! text_posn
          text_rotate1 = Geom::Transformation.rotation text_center, Z_AXIS , @text_rotn
          text_group.transform! text_rotate1

        #=========================
        else  # we're dimensioning an exterior angle

          # Draw arcs
          dim_will_fit_text_width = true
          # Scaling arc segments to the angle gives smoother arc
          segments = (self.class.arc_segments *  (complement/90.degrees)).to_i
          if @quadrant == "northerly" || @quadrant == "southerly"
            arc1 = ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius, 0, - 0.5*(complement - text_gap_angle[0]), segments
            arc2 = ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius, - 0.5*(complement + text_gap_angle[0]),
                                - complement, segments
          else
            arc1 = ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius, 0, - 0.5*(complement - text_gap_angle[1]), segments
            arc2 = ents.add_arc ORIGIN, X_AXIS, Z_AXIS, @radius, - 0.5*(complement + text_gap_angle[1]),
                                - complement, segments
          end
          # Draw arrowheads
          arrow1_rotn = -90.degrees
          arrow2_rotn = 90.degrees + angle

          arrow1_rotate = Geom::Transformation.rotation(ORIGIN, Z_AXIS, arrow1_rotn)
          arrow2_rotate = Geom::Transformation.rotation(ORIGIN, Z_AXIS, arrow2_rotn)
          arrow1_move = Geom::Transformation.translation(ORIGIN.vector_to(arc1[0].start.position))
          arrow2_move = Geom::Transformation.translation(ORIGIN.vector_to(arc2[-1].end.position))

          # Combine transformations to insert an arrowhead at start and end of arcs

          dim_angle_arrow = get_arrowhead_definition(self.class.arrow_style)

          arrow1 = ents.add_instance dim_angle_arrow, arrow1_move*arrow1_rotate*arrow_size_scale
          arrow1 = ents.add_instance dim_angle_arrow, arrow2_move*arrow2_rotate*arrow_size_scale

          #  Move the center of the text to the center of the dimension arc ...
          text_center = Geom::Point3d.new [@radius*Math::cos(0.5*complement), -@radius*Math::sin(0.5*complement),0]
          text_posn = text_center - text_bb_center + [0, 0, 0.001*@radius]
          text_group.move! text_posn
          # ... and rotate it in line with middle of arc
          text_rotate1 = Geom::Transformation.rotation text_center, Z_AXIS , @text_rotn
          text_group.transform! text_rotate1
        end # if @inside/else outside angle
      end

      group.transform! two_vec_transformation(vec1, normal, @pts[1])

      # tell undo the end of the bundled operation
      model.commit_operation

      #start over
      @drawn = true
      @state = 0
      show_status
    end # draw_angle_dim

    def two_vec_transformation(vec1, normal, vertex)
      # Construct a Transformation to move and orient the dimension group
      # drawn at the origin to the location and orientation of the
      # angle.
      # vec1 is the "right hand side" vector of the angle, seen looking out
      # from the vertex, i.e. the target orientation of the x axis of the
      # dimension as drawn at the origin
      # normal is the normal to the plane of the angle, i.e. the target
      # orientation of the z axis of the dimension as drawn at the origin.
      # This uses the (poorly documented) representation of a Transformation
      # as a 4x4 array in the SketchUp Ruby API.

      newx = vec1.normalize
      newz = normal
      newy = newz.cross(newx).normalize

      # use the vectors to build a transformation matrix
      trans_array = []
      trans_array[0] = newx[0]
      trans_array[1] = newx[1]
      trans_array[2] = newx[2]
      trans_array[3] = 0.0
      trans_array[4] = newy[0]
      trans_array[5] = newy[1]
      trans_array[6] = newy[2]
      trans_array[7] = 0.0
      trans_array[8] = newz[0]
      trans_array[9] = newz[1]
      trans_array[10] = newz[2]
      trans_array[11] = 0.0
      trans_array[12] = vertex[0]
      trans_array[13] = vertex[1]
      trans_array[14] = vertex[2]
      trans_array[15] = 1.0

      trans_from_array = Geom::Transformation.new trans_array

    end
    # ---------------------------------------------------------------------------------------------

    def which_quadrant(vec)
      # Check orientation of angle bisector direction relative to vertex, in camera's view.
      # Returned values from view_screen_coords are X from left to right (in pixels, approximately) and
      #   Y values measured from TOP to BOTTOM (reverse of normal Y axis direction)
      view = Sketchup.active_model.active_view
      pt_vertex = view.screen_coords @pts[1]

      pt_text_posn = view.screen_coords vec[1]

      # Now calculate which quadrant the vertex-to-text-centre line appears to be in the current view
      #   and return the quadrant as a string value

      diff_x = (pt_text_posn.x - pt_vertex.x)
      diff_y = (pt_text_posn.y - pt_vertex.y)

      if diff_y < 0 && diff_y.abs > diff_x.abs
        # vertex to text center points in northerly quadrant (angle between +45  and +135 degrees)
        return "northerly"
      elsif diff_x < 0 && diff_y.abs <= diff_x.abs
        # vertex to text center points in westerly quadrant (angle between +135 and -135)
        return "westerly"
      elsif diff_y >= 0 && diff_y.abs > diff_x.abs
        # vertex to text center points in southerly quadrant (angle between -45 and -135)
        return "southerly"
      elsif diff_x >= 0 && diff_y.abs <= diff_x.abs
        #  vertex to text center points in easterly quadrant (angle between +135 and -135)
        return "easterly"
      end
    end

    #-----------------------------------------------------------------------------
    # advance to the next state and set the status texts appropriately.  Used by tool callbacks.
    def increment_state
      @state += 1
      case @state
      when 1
        # have first pick, ready for second
        # retain InputPoint as a hint for next
        @ip1.copy! @ip
        Sketchup::set_status_text((DangLH['Point 2: vertex of measured angle']))
      when 2
        # have second pick, ready for third
        # retain InputPoint as a hint for next
        @ip1.copy! @ip
        Sketchup::set_status_text((DangLH['Point 3: second end of measured angle']))
      when 3
        # have three picks, ready to draw the angle dimension
        draw_angle_dim
      end
    end # increment_state

  end # class DrawAngleDimTool


end # AngularDim
end # SLB_JWMPlugins
