######
# Message logging facility
#
# To use, create a new instance of Logger.
# By default, the Logger will be active immediately and will write to a
# file named "AD log" in the "log" subfolder of the extension's folder
# (it will create the folder if it doesn't exist).
# State of a Logger can be set via logging=().

module SLB_JWMPlugins
  module AngularDim
    class Logger

      def initialize(init_state = true, log_name = "AD log")
        @logging = init_state
        @logpath = File.join(EXTPATH, "log")
        Dir.mkdir(@logpath) unless File.directory?(@logpath)
        @logfile = File.join(@logpath, log_name)
      end

      def logging=(bool_value)
        # argument must be either the true singleton or the false singleton, not something
        # that just happens to work in true/false logic such as nil or 0.
        @logging = bool_value if (bool_value.is_a?(TrueClass) || bool_value.is_a?(FalseClass))
      end
      def logging
        @logging
      end
      
      def log_message(message)
        if logging
          File.open(@logfile, "a") do |file|
            file.write(message)
            # prevent accidental concatenation of messages on same line
            file.write("\n") unless message[-1] == "\n"
          end
        end
      end
      
    end # Class Logger
  end # module Angular Dim
end # module SLB_JWMPlugins

